"""Parse and format data types, from and to packets that can be sent.

types are automatically inferred using the `PARSERS` and `WRITERS` members.

Allowed types are:
    int (but not *long* ints) -> osc int
    floats -> osc float
    bytes (encoded strings) -> osc strings
    bytearray (raw data) -> osc blob

"""

__all__ = (
    'parse',
    'read_packet', 'read_message', 'read_bundle',
    'format_bundle', 'format_message',
    'MidiTuple',
)


from struct import Struct, pack, unpack_from, calcsize
from time import time
import sys
from collections import Counter, namedtuple
from oscpy.stats import Stats

if sys.version_info.major > 2:  # pragma: no cover
    UNICODE = str
    izip = zip
else:  # pragma: no cover
    UNICODE = unicode
    from itertools import izip

INT = Struct('>i')
FLOAT = Struct('>f')
STRING = Struct('>s')
TIME_TAG = Struct('>II')

TP_PACKET_FORMAT = "!12I"
# 1970-01-01 00:00:00
NTP_DELTA = 2208988800

NULL = b'\0'
EMPTY = tuple()
INF = float('inf')

MidiTuple = namedtuple('MidiTuple', 'port_id status_byte data1 data2')

def padded(l, n=4):
    """Return the size to pad a thing to.

    - `l` being the current size of the thing.
    - `n` being the desired divisor of the thing's padded size.
    """
    return n * (min(1, divmod(l, n)[1]) + l // n)


def parse_int(value, offset=0, **kwargs):
    """Return an int from offset in value."""
    return INT.unpack_from(value, offset)[0], INT.size


def parse_float(value, offset=0, **kwargs):
    """Return a float from offset in value."""
    return FLOAT.unpack_from(value, offset)[0], FLOAT.size


def parse_string(value, offset=0, encoding='', encoding_errors='strict'):
    """Return a string from offset in value.

    If encoding is defined, the string will be decoded. `encoding_errors`
    will be used to manage encoding errors in decoding.
    """
    result = []
    count = 0
    ss = STRING.size
    while True:
        c = STRING.unpack_from(value, offset + count)[0]
        count += ss

        if c == NULL:
            break
        result.append(c)

    r = b''.join(result)
    if encoding:
        return r.decode(encoding, errors=encoding_errors), padded(count)
    else:
        return r, padded(count)


def parse_blob(value, offset=0, **kwargs):
    """Return a blob from offset in value."""
    size = calcsize('>i')
    length = unpack_from('>i', value, offset)[0]
    data = unpack_from('>%iQ' % length, value, offset + size)
    return data, padded(length, 8)


def parse_midi(value, offset=0, **kwargs):
    """Return a MIDI tuple from offset in value.
    A valid MIDI message: (port id, status byte, data1, data2).
    """
    val = unpack_from('>I', value, offset)[0]
    args = tuple((val & 0xFF << 8 * i) >> 8 * i for i in range(3, -1, -1))
    midi = MidiTuple(*args)
    return midi, len(midi)


def format_midi(value):
    return sum((val & 0xFF) << 8 * (3 - pos) for pos, val in enumerate(value))


def parse_true(*args, **kwargs):
    return True, 0


def format_true(value):
    return EMPTY


def parse_false(*args, **kwargs):
    return False, 0


def format_false(value):
    return EMPTY


def parse_nil(*args, **kwargs):
    return None, 0


def format_nil(value):
    return EMPTY


def parse_infinitum(*args, **kwargs):
    return INF, 0


def format_infinitum(value):
    return EMPTY


PARSERS = {
    b'i': parse_int,
    b'f': parse_float,
    b's': parse_string,
    b'S': parse_string,
    b'b': parse_blob,
    b'm': parse_midi,
    b'T': parse_true,
    b'F': parse_false,
    b'N': parse_nil,
    b'I': parse_infinitum,
    # TODO
    # b'h': parse_long,
    # b't': parse_timetage,
    # b'd': parse_double,
    # b'c': parse_char,
    # b'r': parse_rgba,
    # b'[': parse_array_start,
    # b']': parse_array_end,
}


PARSERS.update({
    ord(k): v
    for k, v in PARSERS.items()
})


WRITERS = (
    (float, (b'f', b'f')),
    (int, (b'i', b'i')),
    (bytes, (b's', b'%is')),
    (UNICODE, (b's', b'%is')),
    (bytearray, (b'b', b'%ib')),
    (True, (b'T', b'')),
    (False, (b'F', b'')),
    (None, (b'N', b'')),
    (MidiTuple, (b'm', b'I')),
)


PADSIZES = {
    bytes: 4,
    bytearray: 8
}


def parse(hint, value, offset=0, encoding='', encoding_errors='strict'):
    """Call the correct parser function for the provided hint.

    `hint` will be used to determine the correct parser, other parameters
    will be passed to this parser.
    """
    parser = PARSERS.get(hint)

    if not parser:
        raise ValueError(
            "no known parser for type hint: {}, value: {}".format(hint, value)
        )

    return parser(
        value, offset=offset, encoding=encoding,
        encoding_errors=encoding_errors
    )


def format_message(address, values, encoding='', encoding_errors='strict'):
    """Create a message."""
    tags = [b',']
    fmt = []

    encode_cache = {}

    lv = 0
    count = Counter()

    for value in values:
        lv += 1
        cls_or_value, writer = None, None
        for cls_or_value, writer in WRITERS:
            if (
                cls_or_value is value
                or isinstance(cls_or_value, type)
                and isinstance(value, cls_or_value)
            ):
                break
        else:
            raise TypeError(
                u'unable to find a writer for value {}, type not in: {}.'
                .format(value, [x[0] for x in WRITERS])
            )

        if cls_or_value == UNICODE:
            if not encoding:
                raise TypeError(u"Can't format unicode string without encoding")

            cls_or_value = bytes
            value = (
                encode_cache[value]
                if value in encode_cache else
                encode_cache.setdefault(
                    value, value.encode(encoding, errors=encoding_errors)
                )
            )

        assert cls_or_value, writer

        tag, v_fmt = writer
        if b'%i' in v_fmt:
            v_fmt = v_fmt % padded(len(value) + 1, PADSIZES[cls_or_value])

        tags.append(tag)
        fmt.append(v_fmt)
        count[tag.decode('utf8')] += 1

    fmt = b''.join(fmt)
    tags = b''.join(tags + [NULL])

    if encoding and isinstance(address, UNICODE):
        address = address.encode(encoding, errors=encoding_errors)

    if not address.endswith(NULL):
        address += NULL

    fmt = b'>%is%is%s' % (padded(len(address)), padded(len(tags)), fmt)
    message = pack(
        fmt,
        address,
        tags,
        *(
            (
                encode_cache.get(v) + NULL if isinstance(v, UNICODE) and encoding
                else (v + NULL) if t in (b's', b'b')
                else format_midi(v) if isinstance(v, MidiTuple)
                else v
            )
            for t, v in
            izip(tags[1:], values)
        )
    )
    return message, Stats(1, len(message), lv, count)


def read_message(data, offset=0, encoding='', encoding_errors='strict'):
    """Return address, tags, values, and length of a decoded message.

    Can be called either on a standalone message, or on a message
    extracted from a bundle.
    """
    address, size = parse_string(data, offset=offset)
    index = size
    if not address.startswith(b'/'):
        raise ValueError("address {} doesn't start with a '/'".format(address))

    tags, size = parse_string(data, offset=offset + index)
    if not tags.startswith(b','):
        raise ValueError("tag string {} doesn't start with a ','".format(tags))
    tags = tags[1:]

    index += size

    values = []
    for tag in tags:
        value, off = parse(
            tag, data, offset=offset + index, encoding=encoding,
            encoding_errors=encoding_errors
        )
        values.append(value)
        index += off

    return address, tags, values, index


def time_to_timetag(value):
    """Create a timetag from a time.

    `time` is an unix timestamp (number of seconds since 1/1/1970).
    result is the equivalent time using the NTP format.
    """
    if value is None:
        return (0, 1)
    seconds, fract = divmod(value, 1)
    seconds += NTP_DELTA
    seconds = int(seconds)
    fract = int(fract * 2**32)
    return (seconds, fract)


def timetag_to_time(timetag):
    """Decode a timetag to a time.

    `timetag` is an NTP formated time.
    retult is the equivalent unix timestamp (number of seconds since 1/1/1970).
    """
    if timetag == (0, 1):
        return time()

    seconds, fract = timetag
    return seconds + fract / 2. ** 32 - NTP_DELTA


def format_bundle(data, timetag=None, encoding='', encoding_errors='strict'):
    """Create a bundle from a list of (address, values) tuples.

    String values will be encoded using `encoding` or must be provided
    as bytes.
    `encoding_errors` will be used to manage encoding errors.
    """
    timetag = time_to_timetag(timetag)
    bundle = [pack('8s', b'#bundle\0')]
    bundle.append(TIME_TAG.pack(*timetag))

    stats = Stats()
    for address, values in data:
        msg, st = format_message(
            address, values, encoding='',
            encoding_errors=encoding_errors
        )
        bundle.append(pack('>i', len(msg)))
        bundle.append(msg)
        stats += st

    return b''.join(bundle), stats


def read_bundle(data, encoding='', encoding_errors='strict'):
    """Decode a bundle into a (timestamp, messages) tuple."""
    length = len(data)

    header = unpack_from('7s', data, 0)[0]
    offset = 8 * STRING.size
    if header != b'#bundle':
        raise ValueError(
            "the message doesn't start with '#bundle': {}".format(header))

    timetag = timetag_to_time(TIME_TAG.unpack_from(data, offset))
    offset += TIME_TAG.size

    messages = []
    while offset < length:
        # NOTE, we don't really care about the size of the message, our
        # parsing will compute it anyway
        # size = Int.unpack_from(data, offset)
        offset += INT.size
        address, tags, values, off = read_message(
            data, offset, encoding=encoding, encoding_errors=encoding_errors
        )
        offset += off
        messages.append((address, tags, values, offset))

    return (timetag, messages)


def read_packet(data, drop_late=False, encoding='', encoding_errors='strict'):
    """Detect if the data received is a simple message or a bundle, read it.

    Always return a list of messages.
    If drop_late is true, and the received data is an expired bundle,
    then returns an empty list.
    """
    header = unpack_from('>c', data, 0)[0]
    if header == b'/':
        return [
            read_message(
                data, encoding=encoding,
                encoding_errors=encoding_errors
            )
        ]

    elif header == b'#':
        timetag, messages = read_bundle(
            data, encoding=encoding, encoding_errors=encoding_errors
        )
        if drop_late:
            if time() > timetag:
                return []
        return messages
    else:
        raise ValueError('packet is not a message or a bundle')



"""Server API.

This module currently only implements `OSCThreadServer`, a thread based server.
"""
import logging
from threading import Thread, Event

import os
import re
import inspect
from sys import platform
from time import sleep, time
from functools import partial
from select import select
import socket

from oscpy import __version__
from oscpy.parser import read_packet, UNICODE
from oscpy.client import send_bundle, send_message
from oscpy.stats import Stats


logger = logging.getLogger(__name__)


def ServerClass(cls):
    """Decorate classes with for methods implementing OSC endpoints.

    This decorator is necessary on your class if you want to use the
    `address_method` decorator on its methods, see
    `:meth:OSCThreadServer.address_method`'s documentation.
    """
    cls_init = cls.__init__

    def __init__(self, *args, **kwargs):
        cls_init(self, *args, **kwargs)

        for m in dir(self):
            meth = getattr(self, m)
            if hasattr(meth, '_address'):
                server, address, sock, get_address = meth._address
                server.bind(address, meth, sock, get_address=get_address)

    cls.__init__ = __init__
    return cls


__FILE__ = inspect.getfile(ServerClass)


class OSCThreadServer(object):
    """A thread-based OSC server.

    Listens for osc messages in a thread, and dispatches the messages
    values to callbacks from there.

    The '/_oscpy/' namespace is reserved for metadata about the OSCPy
    internals, please see package documentation for further details.
    """

    def __init__(
        self, drop_late_bundles=False, timeout=0.01, advanced_matching=False,
        encoding='', encoding_errors='strict', default_handler=None, intercept_errors=True
    ):
        """Create an OSCThreadServer.

        - `timeout` is a number of seconds used as a time limit for
          select() calls in the listening thread, optiomal, defaults to
          0.01.
        - `drop_late_bundles` instruct the server not to dispatch calls
          from bundles that arrived after their timetag value.
          (optional, defaults to False)
        - `advanced_matching` (defaults to False), setting this to True
          activates the pattern matching part of the specification, let
          this to False if you don't need it, as it triggers a lot more
          computation for each received message.
        - `encoding` if defined, will be used to encode/decode all
          strings sent/received to/from unicode/string objects, if left
          empty, the interface will only accept bytes and return bytes
          to callback functions.
        - `encoding_errors` if `encoding` is set, this value will be
          used as `errors` parameter in encode/decode calls.
        - `default_handler` if defined, will be used to handle any
          message that no configured address matched, the received
          arguments will be (address, *values).
        - `intercept_errors`, if True, means that exception raised by
          callbacks will be intercepted and logged. If False, the handler
          thread will terminate mostly silently on such exceptions.
        """
        self._must_loop = True
        self._termination_event = Event()

        self.addresses = {}
        self.sockets = []
        self.timeout = timeout
        self.default_socket = None
        self.drop_late_bundles = drop_late_bundles
        self.advanced_matching = advanced_matching
        self.encoding = encoding
        self.encoding_errors = encoding_errors
        self.default_handler = default_handler
        self.intercept_errors = intercept_errors

        self.stats_received = Stats()
        self.stats_sent = Stats()

        t = Thread(target=self._run_listener)
        t.daemon = True
        t.start()
        self._thread = t

        self._smart_address_cache = {}
        self._smart_part_cache = {}

    def bind(self, address, callback, sock=None, get_address=False):
        """Bind a callback to an osc address.

        A socket in the list of existing sockets of the server can be
        given. If no socket is provided, the default socket of the
        server is used, if no default socket has been defined, a
        RuntimeError is raised.

        Multiple callbacks can be bound to the same address.
        """
        if not sock and self.default_socket:
            sock = self.default_socket
        elif not sock:
            raise RuntimeError('no default socket yet and no socket provided')

        if isinstance(address, UNICODE) and self.encoding:
            address = address.encode(
                self.encoding, errors=self.encoding_errors)

        if self.advanced_matching:
            address = self.create_smart_address(address)

        callbacks = self.addresses.get((sock, address), [])
        cb = (callback, get_address)
        if cb not in callbacks:
            callbacks.append(cb)
        self.addresses[(sock, address)] = callbacks

    def create_smart_address(self, address):
        """Create an advanced matching address from a string.

        The address will be split by '/' and each part will be converted
        into a regexp, using the rules defined in the OSC specification.
        """
        cache = self._smart_address_cache

        if address in cache:
            return cache[address]

        else:
            parts = address.split(b'/')
            smart_parts = tuple(
                re.compile(self._convert_part_to_regex(part)) for part in parts
            )
            cache[address] = smart_parts
            return smart_parts

    def _convert_part_to_regex(self, part):
        cache = self._smart_part_cache

        if part in cache:
            return cache[part]

        else:
            r = [b'^']
            for i, _ in enumerate(part):
                # getting a 1 char byte string instead of an int in
                # python3
                c = part[i:i + 1]
                if c == b'?':
                    r.append(b'.')
                elif c == b'*':
                    r.append(b'.*')
                elif c == b'[':
                    r.append(b'[')
                elif c == b'!' and r and r[-1] == b'[':
                    r.append(b'^')
                elif c == b']':
                    r.append(b']')
                elif c == b'{':
                    r.append(b'(')
                elif c == b',':
                    r.append(b'|')
                elif c == b'}':
                    r.append(b')')
                else:
                    r.append(c)

            r.append(b'$')

            smart_part = re.compile(b''.join(r))

            cache[part] = smart_part
            return smart_part

    def unbind(self, address, callback, sock=None):
        """Unbind a callback from an OSC address.

        See `bind` for `sock` documentation.
        """
        if not sock and self.default_socket:
            sock = self.default_socket
        elif not sock:
            raise RuntimeError('no default socket yet and no socket provided')

        if isinstance(address, UNICODE) and self.encoding:
            address = address.encode(
                self.encoding, errors=self.encoding_errors)

        callbacks = self.addresses.get((sock, address), [])
        to_remove = []
        for cb in callbacks:
            if cb[0] == callback:
                to_remove.append(cb)

        while to_remove:
            callbacks.remove(to_remove.pop())

        self.addresses[(sock, address)] = callbacks

    def listen(
        self, address='localhost', port=0, default=False, family='inet'
    ):
        """Start listening on an (address, port).

        - if `port` is 0, the system will allocate a free port
        - if `default` is True, the instance will save this socket as the
          default one for subsequent calls to methods with an optional socket
        - `family` accepts the 'unix' and 'inet' values, a socket of the
          corresponding type will be created.
          If family is 'unix', then the address must be a filename, the
          `port` value won't be used. 'unix' sockets are not defined on
          Windows.

        The socket created to listen is returned, and can be used later
        with methods accepting the `sock` parameter.
        """
        if family == 'unix':
            family_ = socket.AF_UNIX
        elif family == 'inet':
            family_ = socket.AF_INET
        else:
            raise ValueError(
                "Unknown socket family, accepted values are 'unix' and 'inet'"
            )

        sock = socket.socket(family_, socket.SOCK_DGRAM)
        if family == 'unix':
            addr = address
        else:
            addr = (address, port)
        sock.bind(addr)
        self.sockets.append(sock)
        if default and not self.default_socket:
            self.default_socket = sock
        elif default:
            raise RuntimeError(
                'Only one default socket authorized! Please set '
                'default=False to other calls to listen()'
            )
        self.bind_meta_routes(sock)
        return sock

    def close(self, sock=None):
        """Close a socket opened by the server."""
        if not sock and self.default_socket:
            sock = self.default_socket
        elif not sock:
            raise RuntimeError('no default socket yet and no socket provided')

        if platform != 'win32' and sock.family == socket.AF_UNIX:
            os.unlink(sock.getsockname())
        else:
            sock.close()

        if sock == self.default_socket:
            self.default_socket = None

    def getaddress(self, sock=None):
        """Wrap call to getsockname.

        If `sock` is None, uses the default socket for the server.

        Returns (ip, port) for an inet socket, or filename for an unix
        socket.
        """
        if not sock and self.default_socket:
            sock = self.default_socket
        elif not sock:
            raise RuntimeError('no default socket yet and no socket provided')

        return sock.getsockname()

    def stop(self, s=None):
        """Close and remove a socket from the server's sockets.

        If `sock` is None, uses the default socket for the server.

        """
        if not s and self.default_socket:
            s = self.default_socket

        if s in self.sockets:
            read = select([s], [], [], 0)
            s.close()
            if s in read:
                s.recvfrom(65535)
            self.sockets.remove(s)
        else:
            raise RuntimeError('{} is not one of my sockets!'.format(s))

    def stop_all(self):
        """Call stop on all the existing sockets."""
        for s in self.sockets[:]:
            self.stop(s)
        sleep(10e-9)

    def terminate_server(self):
        """Request the inner thread to finish its tasks and exit.

        May be called from an event, too.
        """
        self._must_loop = False

    def join_server(self, timeout=None):
        """Wait for the server to exit (`terminate_server()` must have been called before).

        Returns True if and only if the inner thread exited before timeout."""
        return self._termination_event.wait(timeout=timeout)

    def _run_listener(self):
        """Wrapper just ensuring that the handler thread cleans up on exit."""
        try:
            self._listen()
        finally:
            self._termination_event.set()

    def _listen(self):
        """(internal) Busy loop to listen for events.

        This method is called in a thread by the `listen` method, and
        will be the one actually listening for messages on the server's
        sockets, and calling the callbacks when messages are received.
        """

        match = self._match_address
        advanced_matching = self.advanced_matching
        addresses = self.addresses
        stats = self.stats_received

        def _execute_callbacks(_callbacks_list):
            for cb, get_address in _callbacks_list:
                try:
                    if get_address:
                        cb(address, *values)
                    else:
                        cb(*values)
                except Exception as exc:
                    if self.intercept_errors:
                        logger.error("Unhandled exception caught in oscpy server", exc_info=True)
                    else:
                        raise

        while self._must_loop:

            drop_late = self.drop_late_bundles
            if not self.sockets:
                sleep(.01)
                continue
            else:
                try:
                    read, write, error = select(self.sockets, [], [], self.timeout)
                except (ValueError, socket.error):
                    continue

            for sender_socket in read:
                try:
                    data, sender = sender_socket.recvfrom(65535)
                except ConnectionResetError:
                    continue

                for address, tags, values, offset in read_packet(
                    data, drop_late=drop_late, encoding=self.encoding,
                    encoding_errors=self.encoding_errors
                ):
                    stats.calls += 1
                    stats.bytes += offset
                    stats.params += len(values)
                    stats.types.update(tags)

                    matched = False
                    if advanced_matching:
                        for sock, addr in addresses:
                            if sock == sender_socket and match(addr, address):
                                callbacks_list = addresses.get((sock, addr), [])
                                if callbacks_list:
                                    matched = True
                                    _execute_callbacks(callbacks_list)
                    else:
                        callbacks_list = addresses.get((sender_socket, address), [])
                        if callbacks_list:
                            matched = True
                            _execute_callbacks(callbacks_list)

                    if not matched and self.default_handler:
                        self.default_handler(address, *values)

    @staticmethod
    def _match_address(smart_address, target_address):
        """(internal) Check if provided `smart_address` matches address.

        A `smart_address` is a list of regexps to match
        against the parts of the `target_address`.
        """
        target_parts = target_address.split(b'/')
        if len(target_parts) != len(smart_address):
            return False

        return all(
            model.match(part)
            for model, part in
            zip(smart_address, target_parts)
        )

    def send_message(
        self, osc_address, values, ip_address, port, sock=None, safer=False
    ):
        """Shortcut to the client's `send_message` method.

        Use the default_socket of the server by default.
        See `client.send_message` for more info about the parameters.
        """
        if not sock and self.default_socket:
            sock = self.default_socket
        elif not sock:
            raise RuntimeError('no default socket yet and no socket provided')

        stats = send_message(
            osc_address,
            values,
            ip_address,
            port,
            sock=sock,
            safer=safer,
            encoding=self.encoding,
            encoding_errors=self.encoding_errors
        )
        self.stats_sent += stats
        return stats

    def send_bundle(
        self, messages, ip_address, port, timetag=None, sock=None, safer=False
    ):
        """Shortcut to the client's `send_bundle` method.

        Use the `default_socket` of the server by default.
        See `client.send_bundle` for more info about the parameters.
        """
        if not sock and self.default_socket:
            sock = self.default_socket
        elif not sock:
            raise RuntimeError('no default socket yet and no socket provided')

        stats = send_bundle(
            messages,
            ip_address,
            port,
            sock=sock,
            safer=safer,
            encoding=self.encoding,
            encoding_errors=self.encoding_errors
        )
        self.stats_sent += stats
        return stats

    def get_sender(self):
        """Return the socket, ip and port of the message that is currently being managed.
        Warning::

            this method should only be called from inside the handling
            of a message (i.e, inside a callback).
        """
        frames = inspect.getouterframes(inspect.currentframe())
        for frame, filename, _, function, _, _ in frames:
            if function == '_listen' and __FILE__.startswith(filename):
                break
        else:
            raise RuntimeError('get_sender() not called from a callback')

        sock = frame.f_locals.get('sender_socket')
        address, port = frame.f_locals.get('sender')
        return sock, address, port

    def answer(
        self, address=None, values=None, bundle=None, timetag=None,
        safer=False, port=None
    ):
        """Answers a message or bundle to a client.

        This method can only be called from a callback, it will lookup
        the sender of the packet that triggered the callback, and send
        the given message or bundle to it.

        `timetag` is only used if `bundle` is True.
        See `send_message` and `send_bundle` for info about the parameters.

        Only one of `values` or `bundle` should be defined, if `values`
        is defined, `send_message` is used with it, if `bundle` is
        defined, `send_bundle` is used with its value.
        """
        if not values:
            values = []

        sock, ip_address, response_port = self.get_sender()

        if port is not None:
            response_port = port

        if bundle:
            return self.send_bundle(
                bundle, ip_address, response_port, timetag=timetag, sock=sock,
                safer=safer
            )
        else:
            return self.send_message(
                address, values, ip_address, response_port, sock=sock
            )

    def address(self, address, sock=None, get_address=False):
        """Decorate functions to bind them from their definition.

        `address` is the osc address to bind to the callback.
        if `get_address` is set to True, the first parameter the
        callback will receive will be the address that matched (useful
        with advanced matching).

        example:
            server = OSCThreadServer()
            server.listen('localhost', 8000, default=True)

            @server.address(b'/printer')
            def printer(values):
                print(values)

            send_message(b'/printer', [b'hello world'])

        note:
            This won't work on methods as it'll call them as normal
            functions, and the callback won't get a `self` argument.

            To bind a method use the `address_method` decorator.
        """
        def decorator(callback):
            self.bind(address, callback, sock, get_address=get_address)
            return callback

        return decorator

    def address_method(self, address, sock=None, get_address=False):
        """Decorate methods to bind them from their definition.

        The class defining the method must itself be decorated with the
        `ServerClass` decorator, the methods will be bound to the
        address when the class is instantiated.

        See `address` for more information about the parameters.

        example:

            osc = OSCThreadServer()
            osc.listen(default=True)

            @ServerClass
            class MyServer(object):

                @osc.address_method(b'/test')
                def success(self, *args):
                    print("success!", args)
        """
        def decorator(decorated):
            decorated._address = (self, address, sock, get_address)
            return decorated

        return decorator

    def bind_meta_routes(self, sock=None):
        """This module implements osc routes to probe the internal state of a
        live OSCPy server. These routes are placed in the /_oscpy/ namespace,
        and provide information such as the version, the existing routes, and
        usage statistics of the server over time.

        These requests will be sent back to the client's address/port that sent
        them, with the osc address suffixed with '/answer'.

        examples:
            '/_oscpy/version' -> '/_oscpy/version/answer'
            '/_oscpy/stats/received' -> '/_oscpy/stats/received/answer'

        messages to these routes require a port number as argument, to
        know to which port to send to.
        """
        self.bind(b'/_oscpy/version', self._get_version, sock=sock)
        self.bind(b'/_oscpy/routes', self._get_routes, sock=sock)
        self.bind(b'/_oscpy/stats/received', self._get_stats_received, sock=sock)
        self.bind(b'/_oscpy/stats/sent', self._get_stats_sent, sock=sock)

    def _get_version(self, port, *args):
        self.answer(
            b'/_oscpy/version/answer',
            (__version__, ),
            port=port
        )

    def _get_routes(self, port, *args):
        self.answer(
            b'/_oscpy/routes/answer',
            [a[1] for a in self.addresses],
            port=port
        )

    def _get_stats_received(self, port, *args):
        self.answer(
            b'/_oscpy/stats/received/answer',
            self.stats_received.to_tuple(),
            port=port
        )

    def _get_stats_sent(self, port, *args):
        self.answer(
            b'/_oscpy/stats/sent/answer',
            self.stats_sent.to_tuple(),
            port=port
        )



"""Client API.

This module provides both a functional and an object oriented API.

You can use directly `send_message`, `send_bundle` and the `SOCK` socket
that is created by default, or use `OSCClient` to store parameters common
to your requests and avoid repeating them in your code.
"""

import socket
from time import sleep
from sys import platform

from oscpy.parser import format_message, format_bundle
from oscpy.stats import Stats

SOCK = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)


def send_message(
    osc_address, values, ip_address, port, sock=SOCK, safer=False,
    encoding='', encoding_errors='strict'
):
    """Send an osc message to a socket address.

    - `osc_address` is the osc endpoint to send the data to (e.g b'/test')
      it should be a bytestring
    - `values` is the list of values to send, they can be any supported osc
      type (bytestring, float, int, blob...)
    - `ip_address` can either be an ip address if the used socket is of
      the AF_INET family, or a filename if the socket is of type AF_UNIX
    - `port` value will be ignored if socket is of type AF_UNIX
    - `sock` should be a socket object, the client's default socket can be
      used as default
    - the `safer` parameter allows to wait a little after sending, to make
      sure the message is actually sent before doing anything else,
      should only be useful in tight loop or cpu-busy code.
    - `encoding` if defined, will be used to encode/decode all
      strings sent/received to/from unicode/string objects, if left
      empty, the interface will only accept bytes and return bytes
      to callback functions.
    - `encoding_errors` if `encoding` is set, this value will be
      used as `errors` parameter in encode/decode calls.

    examples:
        send_message(b'/test', [b'hello', 1000, 1.234], 'localhost', 8000)
        send_message(b'/test', [], '192.168.0.1', 8000, safer=True)

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        send_message(b'/test', [], '192.168.0.1', 8000, sock=sock, safer=True)

        # unix sockets work on linux and osx, and over unix platforms,
        # but not windows
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
        send_message(b'/some/address', [1, 2, 3], b'/tmp/sock')

    """
    if platform != 'win32' and sock.family == socket.AF_UNIX:
        address = ip_address
    else:
        address = (ip_address, port)

    message, stats = format_message(
        osc_address, values, encoding=encoding,
        encoding_errors=encoding_errors
    )

    sock.sendto(message, address)
    if safer:
        sleep(10e-9)

    return stats


def send_bundle(
    messages, ip_address, port, timetag=None, sock=None, safer=False,
    encoding='', encoding_errors='strict'
):
    """Send a bundle built from the `messages` iterable.

    each item in the `messages` list should be a two-tuple of the form:
    (address, values).

    example:
        (
            ('/create', ['name', 'value']),
            ('/select', ['name']),
            ('/update', ['name', 'value2']),
            ('/delete', ['name']),
        )

    `timetag` is optional but can be a float of the number of seconds
    since 1970 when the events described in the bundle should happen.

    See `send_message` documentation for the other parameters.
    """
    if not sock:
        sock = SOCK
    bundle, stats = format_bundle(
        messages, timetag=timetag, encoding=encoding,
        encoding_errors=encoding_errors
    )
    sock.sendto(bundle, (ip_address, port))
    if safer:
        sleep(10e-9)

    return stats


class OSCClient(object):
    """Class wrapper for the send_message and send_bundle functions.

    Allows to define `address`, `port` and `sock` parameters for all calls.
    If encoding is provided, all string values will be encoded
    into this encoding before being sent.
    """

    def __init__(
        self, address, port, sock=None, encoding='', encoding_errors='strict'
    ):
        """Create an OSCClient.

        `address` and `port` are the destination of messages sent
        by this client. See `send_message` and `send_bundle` documentation
        for more information.
        """
        self.address = address
        self.port = port
        self.sock = sock or SOCK
        self.encoding = encoding
        self.encoding_errors = encoding_errors
        self.stats = Stats()

    def send_message(self, address, values, safer=False):
        """Wrap the module level `send_message` function."""
        stats = send_message(
            address, values, self.address, self.port, self.sock,
            safer=safer, encoding=self.encoding,
            encoding_errors=self.encoding_errors
        )
        self.stats += stats
        return stats

    def send_bundle(self, messages, timetag=None, safer=False):
        """Wrap the module level `send_bundle` function."""
        stats = send_bundle(
            messages, self.address, self.port, timetag=timetag,
            sock=self.sock, safer=safer, encoding=self.encoding,
            encoding_errors=self.encoding_errors
        )
        self.stats += stats
        return stats



"Simple utility class to gather stats about the volumes of data managed"

from collections import Counter


class Stats(object):
    def __init__(self, calls=0, bytes=0, params=0, types=None, **kwargs):
        self.calls = calls
        self.bytes = bytes
        self.params = params
        self.types = types or Counter()
        super(Stats, self).__init__(**kwargs)

    def to_tuple(self):
        types = self.types
        keys = types.keys()
        return (
            self.calls,
            self.bytes,
            self.params,
            ''.join(keys),
        ) + tuple(types[k] for k in keys)

    def __iadd__(self, other):
        assert isinstance(other, Stats)
        self.calls += other.calls
        self.bytes += other.bytes
        self.params += other.params
        self.types += other.types
        return self

    def __add__(self, other):
        assert isinstance(other, Stats)
        return Stats(
            calls=self.calls + other.calls,
            bytes=self.bytes + other.bytes,
            params=self.params + other.params,
            types=self.types + other.types
        )

    def __eq__(self, other):
        return other is self or (
            isinstance(other, Stats)
            and self.calls == self.calls
            and self.bytes == self.bytes
            and self.params == self.params
            and self.types == other.types
        )

    def __repr__(self):
        return 'Stats:\n' + '\n'.join(
            '    {}:{}{}'.format(
                k,
                '' if isinstance(v, str) and v.startswith('\n') else ' ',
                v
            )
            for k, v in (
                ('calls', self.calls),
                ('bytes', self.bytes),
                ('params', self.params),
                (
                    'types',
                    ''.join(
                        '\n        {}: {}'.format(k, self.types[k])
                        for k in sorted(self.types)
                    )
                )
            )
        )



# coding: utf8
"""OSCPy command line tools"""

from argparse import ArgumentParser
from time import sleep
from sys import exit, stderr
from ast import literal_eval

from oscpy.client import send_message
from oscpy.server import OSCThreadServer
from oscpy.stats import Stats


def _send(options):
    def _parse(s):
        try:
            return literal_eval(s)
        except:
            return s

    stats = Stats()
    for i in range(options.repeat):
        stats += send_message(
            options.address,
            [_parse(x) for x in options.message],
            options.host,
            options.port,
            safer=options.safer,
            encoding=options.encoding,
            encoding_errors=options.encoding_errors
        )
    print(stats)


def __dump(options):
    def dump(address, *values):
        print(u'{}: {}'.format(
            address.decode('utf8'),
            ', '.join(
                '{}'.format(
                    v.decode(options.encoding or 'utf8')
                    if isinstance(v, bytes)
                    else v
                )
                for v in values if values
            )
        ))

    osc = OSCThreadServer(
        encoding=options.encoding,
        encoding_errors=options.encoding_errors,
        default_handler=dump
    )
    osc.listen(
        address=options.host,
        port=options.port,
        default=True
    )
    return osc


def _dump(options): # pragma: no cover
    osc = __dump(options)
    try:
        while True:
            sleep(10)
    finally:
        osc.stop()


def init_parser():
    parser = ArgumentParser(description='OSCPy command line interface')
    parser.set_defaults(func=lambda *x: parser.print_usage(stderr))

    subparser = parser.add_subparsers()

    send = subparser.add_parser('send', help='send an osc message to a server')
    send.set_defaults(func=_send)
    send.add_argument('--host', '-H', action='store', default='localhost',
                      help='host (ip or name) to send message to.')
    send.add_argument('--port', '-P', action='store', type=int, default='8000',
                      help='port to send message to.')
    send.add_argument('--encoding', '-e', action='store', default='utf-8',
                      help='how to encode the strings')
    send.add_argument('--encoding_errors', '-E', action='store', default='replace',
                      help='how to treat string encoding issues')
    send.add_argument('--safer', '-s', action='store_true',
                      help='wait a little after sending message')
    send.add_argument('--repeat', '-r', action='store', type=int, default=1,
                      help='how many times to send the message')

    send.add_argument('address', action='store',
                      help='OSC address to send the message to.')
    send.add_argument('message', nargs='*',
                        help='content of the message, separated by spaces.')

    dump = subparser.add_parser('dump', help='listen for messages and print them')
    dump.set_defaults(func=_dump)
    dump.add_argument('--host', '-H', action='store', default='localhost',
                      help='host (ip or name) to send message to.')
    dump.add_argument('--port', '-P', action='store', type=int, default='8000',
                      help='port to send message to.')
    dump.add_argument('--encoding', '-e', action='store', default='utf-8',
                      help='how to encode the strings')
    dump.add_argument('--encoding_errors', '-E', action='store', default='replace',
                      help='how to treat string encoding issues')

    # bridge = parser.add_parser('bridge', help='listen for messages and redirect them to a server')
    return parser


def main(): # pragma: no cover
    parser = init_parser()
    options = parser.parse_args()
    exit(options.func(options))
